function(input,output,session){



  if(!exists("con")){
    con <- connect_api(api)
  }



  # Rprof(paste0("/tmp/",strftime(Sys.time(), "%Y-%m-%d-%H-%M-%S.Rprof")),
  #       interval = 0.01, line.profiling = TRUE,
  #       gc.profiling = TRUE, memory.profiling = TRUE)
  #
  # onStop(function() {
  #   Rprof(NULL)
  # })


  # ENV ####
  # 1.Sites ####

  sites <- reactive({
    status <- parseQueryString(isolate(session$clientData$url_search))['status'][[1]]
    sites <- get_sites(con)
    if(is.null(status)){return(sites)}
    sites[which(sites$status == status),]
  })

  # 2.Stations ####


  # stations_enu <- reactive({
  #   # get all ENU coordinates for all stations of the site
  #   stations <- stations_list()
  #   dr <- input$date_range
  #   stations_enu <- lapply(stations$station_id,get_sta_displacement,con=con,date_start = as.POSIXct(dr[1]),date_end = as.POSIXct(dr[2])) %>%
  #     lapply(serie2serie_enu) %>% # convert lat/lon to ENU
  #     lapply(select,date,e,n,u)
  #   stations_enu
  # })

  
  stations <- reactive({
    # list of stations and speed / heading attributes for the selected date range
    req(!is.null(input$site_input))
    sites <- isolate(sites())
    
    id <- sites[which(sites$name == input$site_input),'site_id']
    if(!length(id)){
      return()
    }
    stations <- get_stations(con,id,type='gnss')
    
    if(length(stations$lat) == 0){
      return()
    }

    dr <- input$date_range
    t <- as.numeric(input$solution_type)
    
    # stations_enu <- lapply(stations$station_id,get_sta_displacement,con=con,date_start = as.POSIXct(dr[1]),date_end = as.POSIXct(dr[2])) %>%
    #     lapply(serie2serie_enu) %>% # convert lat/lon to ENU
    #     lapply(select,date,e,n,u)
    #
    # # # filter data on input date range
    # # stations_enu <- lapply(stations_enu,filter,between(date,as.POSIXct(dr[1]),as.POSIXct(dr[2])))

    # # compute speed
    # stations$v <- stations_enu %>%
    #   lapply(speed) %>%
    #   unlist()
    #
    # # compute heading
    # stations$heading <- stations_enu %>%
    #   lapply(heading) %>%
    #   unlist()
    
    #Speed 2
    stations_enu <- lapply(stations$station_id,get_serie,con=con,date_start = as.POSIXct(dr[1]),date_end = as.POSIXct(dr[2])) %>%
      lapply(filter,type==t) %>%
      lapply(serie2serie_enu) %>% # convert lat/lon to ENU
      lapply(select,date,e,n,u) %>%
      lapply(function(l){l[which(!duplicated(l$date)),]}) # remove duplicated rows
    
    speed_heading <- stations_enu %>% lapply(speed2) %>% bind_rows
    stations$v <- speed_heading$speed
    stations$heading <- speed_heading$heading

    # compute color for the map
    rgb <- sapply(stations$v,speed2col) %>% col2rgb

    # send the colors to JS function that compute the CSS filter for the marker
    session$sendCustomMessage(type = "color_svg",t(matrix(rgb,nrow = 3)))
    
    stations
  })

  # get the CSS filter from JS
  style_markers_html <- reactive({
    req(stations)
    f <- input$css_filter
    return(paste0(".",stations()$sta_code,"{",as.list(f),"}"))
  })

  # add the CSS filter to stylesheet
  output$style_markers <- renderUI({
    req(stations)
    style_ <- style_markers_html()
    tags$style(HTML(style_))
  })



  # get the selected station
  station_i <- reactive({
    req(input$station_input)
    station_i <- isolate(stations()) %>% filter(sta_code == input$station_input)
    print(station_i)
    station_i
  })


  # 3.Series ####

  serie <- reactive({
    req(!is.null(input$station_input))
    # get the time serie ENU for the selected station
    if(!is.null(input$station_input) && input$station_input != "" && length(station_i()$station_id)){
      dr <- input$date_range
      t <- as.numeric(input$solution_type)

      tryCatch({
        s <- get_serie(con,station_i()$station_id,as.POSIXct(dr[1]),as.POSIXct(dr[2]))
        s <- s[which(s$type==t),]
        if(dim(s)[1]==0) s<-NULL
        return(serie2serie_enu(s[which(!duplicated(s$date)),]))
      },error=function(cond){
        print(cond)
        return(serie2serie_enu(NULL))
      })

    }
  })
  
  serie_type <- reactive({
    # get the time serie ENU for the selected station
    if(!is.null(input$station_input) && input$station_input != "" && length(station_i()$station_id)){
      dr <- input$date_range
      t <- as.numeric(input$solution_type)
      s <- get_serie(con,station_i()$station_id,as.POSIXct(dr[1]),as.POSIXct(dr[2]))
      return(c(length(s[which(s$type==0),]$type),length(s[which(s$type==1),]$type),length(s[which(s$type==2),]$type)))
    }
    return(c(0,0,0))
  })

  serie_heatmap <- reactive({
    req(input$station_input)
    req(station_i)
    req(serie)
    
    if(!is.null(input$station_input) && input$station_input != "" && length(station_i()$station_id)){
      dr <- input$date_range
      t <- as.numeric(input$solution_type)
      if(t==0){
        s_hm <- get_fix_stat_gipsyx(con,station_i()$station_id,as.POSIXct(dr[1]),as.POSIXct(dr[2]))
      }else if(t==1){
        s_hm <- get_fix_stat_rtklib(con,station_i()$station_id,as.POSIXct(dr[1]),as.POSIXct(dr[2]))
      }else{
        s_hm <- data.frame(date = as.Date(serie()$date),value = 1)
      }
      return(s_hm)
    }
  })
  # '--------- ####
  # UI ####
  # 1.Menu ####
  output$menu <- renderUI({

    tagList(
      selectInput('site_input','Site',choices = sites()$name,selected = sites()[1,'name']),
      selectInput('station_input','Station',choices=""),
      checkboxGroupInput("filter_input",
                         "Filtering",
                         choiceNames = c("Raw data","Filtered data with median","Linear fit"),
                         choiceValues = c("1","2","4"), #powers of 2
                         selected = c("1")),
      dateRangeInput('date_range','Period',start = add_with_rollback(now(),months(-6)),end = now()),
      radioButtons("solution_type",label = "Solution",choices = "")

    )
  })
  
  smallest_window = 14
  observeEvent(input$date_range,{
    if(input$date_range[2] < (input$date_range[1] + smallest_window)){
      showModal(modalDialog(title = "Date range too short", "Date range must be of at least 14 days. Start date will be changed accordingly."))
      updateDateRangeInput(session, "date_range", start=input$date_range[2] - smallest_window)
    }
  })


  # 2.Map ####

  rotatedMarker <- #leaflet plugin for marker rotation
    htmlDependency( name = "Leaflet.rotatedMarker" # note: this .js file has to be copied and stored in your current working directory
                    , version = "0.1.2"
                    , src = normalizePath( path = getwd() )
                    , script = "leaflet.rotatedMarker.js"
    )


  output$map <- renderLeaflet({

    scale <- (stations()$v*100)+15
    scale <- sapply(scale,min,50)
    l <- leaflet() %>%
      registerPlugin(rotatedMarker) %>% # add the plugin to the map
      addTiles("https://wxs.ign.fr/choisirgeoportail/geoportail/wmts?REQUEST=GetTile&SERVICE=WMTS&VERSION=1.0.0&LAYER=ORTHOIMAGERY.ORTHOPHOTOS&STYLE=normal&TILEMATRIXSET=PM&FORMAT=image/jpeg&TILEMATRIX={z}&TILEROW={y}&TILECOL={x}") %>%
      fitBounds(min(stations()$lon)-0.001,min(stations()$lat)-0.001,max(stations()$lon)+0.001,max(stations()$lat)+0.001) %>%
      addMarkers(lng = stations()$lon,
                 lat = stations()$lat,
                 label = stations()$sta_code,
                 layerId = stations()$sta_code,
                 icon = makeIcon(iconUrl= "arrow_up.svg",
                                 iconWidth = scale,
                                 iconHeight = scale,
                                 iconAnchorX = 0.5*scale,
                                 iconAnchorY = scale,
                                 className = stations()$sta_code # this is the class name that is used for styling CSS color
                 ),
                 options = markerOptions(rotationAngle = stations()$heading*180/pi))  %>%
      leaflet::addLegend(colors = speed2col(seq(0,0.3,0.05)),
                         labels = seq(0,0.3,0.05)*100,
                         title = "3D Velocity in cm/yr")

    # browser()
    l
  })

  # 3.Calendar Heatmap ####

  output$heatmap <- renderPlotly({

    if(!is.null(serie()) && !any(is.na(serie()$date))){
      serie_heatmap() %>%
        plot_heatmap() %>%
        layout(height=400)
    }
    else{
      validate(need(F,"No Data, please change the period or select another solution type."))
    }
  })

  # '--------- ####
  # OBSERVERS ####

  # 1.Station input ####

  # observeEvent(input$site_input,{
  #   updateSelectInput(session,"station_input",choices = stations()[,'sta_code'])
  #   print(input$station_input)
  # }, ignoreInit = F)

  observe({
    if(length(stations()[,'sta_code'])==1){
      choices <- c(stations()[,'sta_code'])
      names(choices) <- c(stations()[,'sta_code'])
    }else{
      choices <- stations()[,'sta_code']
    }
    updateSelectInput(session,"station_input",choices = choices,selected=choices[1])
  })

  observeEvent(input$station_input,{
    i<-input$station_input
    
    if(length(i) && i!="" && length(station_i())){

      station_i <- station_i()
      leafletProxy("map") %>%
        addPopups(lng = station_i$lon,
                  lat=station_i$lat,
                  popup=paste0("<b>",station_i$sta_code,"</b></br>3D Velocity : ",round(station_i$v*100,1)," cm/an"),
                  layerId="popup_click")
    }
  }, ignoreInit = T)

  
  # Radio button Solution Type
  observe({
    selected = input$solution_type
    choices = c()
    choices[[paste0("DD - ETRF2000 (",serie_type()[2],")")]] = 1
    choices[[paste0("PPP - ITRF2014 (",serie_type()[1],")")]] = 0
    #choices[[paste0("Tachéo - Local (",serie_type()[3],")")]] = 2
    updateRadioButtons(session,"solution_type",choices = choices,
                      selected = selected)
  })
  
  # 2.Map click ####
  observe({ # when marker click, update selectInput station_input
    m<-input$map_marker_click
    updateSelectInput(session,"station_input",selected=m$id)
  })

  # 3.Time Serie ####
  observe({
    b<-input$filter_input
    o<-sum(as.integer(b))
    if(is.null(b) | is.null(serie())){
      output$ts_2d <- renderPlotly({validate(need(F,"No Data, please change the period or select another solution type."))})
      output$ts<- renderPlotly({validate(need(F,"No Data, please change the period or select another solution type."))})
    }
    else if(any(is.na(serie()$date))){
      output$ts_2d <- renderPlotly({validate(need(F,"No Data, please change the period or select another solution type."))})
      output$ts<- renderPlotly({validate(need(F,"No Data, please change the period or select another solution type."))})
    }
    else{
      output$ts_2d <- renderPlotly({
        ggplot_serie_option_2d(serie(),o)%>%
          layout(title=input$station_input,margin=0.01)
      })
      output$ts <- renderPlotly({
        ggplot_serie_option(serie(),o) %>%
          layout(title=input$station_input,margin=0.01)
      })
    }
  })

  

}
